package jettyTiddlywiki.parser;

import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import jettyTiddlywiki.parser.Tsr.Els.Range;
import jettyTiddlywiki.toolbox.Miscellaneous;
import jettyTiddlywiki.factory.DBF;
import jettyTiddlywiki.factory.TFF;
import jettyTiddlywiki.factory.URCF;

/**
 * recursive tiddlywiki -> dom (without dom4j)
 * 
 * @author debianunsmart
 * 
 */
public class Tsr
{

    interface El
    {
        void fallback(org.w3c.dom.Element e, String s);
        void handle(org.w3c.dom.Element e, String s, Range baseRange);
        void parse(org.w3c.dom.Element e, String s);   
        Range[] match(String s);
    }

    enum Els implements El
    {
        
        a("\\[\\[(.*?)\\]\\]", 1)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 2));
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                em.setAttribute("href", inner.contains("|") ? getPropUrl(inner.split("\\|")[1]) : getPropUrl(inner));
                em.setTextContent(inner.contains("|") ? inner.split("\\|")[0] : inner);
                e.appendChild(em);
                parse(e, s.substring(r.epos + 2));
            }

            private String getPropUrl(final String str) //being used by imageLink
            {
                if (str.startsWith("http") || str.startsWith("ftp") || str.startsWith("mms"))
                {
                    return str;//already url encoded
                }else if(str.startsWith("/") || str.startsWith(".")){
                	return URCF.encode(str);
                }
                return "/wiki/" + URCF.encode(str);
            }
        },
        blockquote("(^|\n)<<<\n([\\s\\S]*?)\n<<<(\n|$)", 2)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 4));
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(em, inner);
                parse(e, s.substring(s.indexOf("<<<", r.epos) + 3));
            }
        },
        left("(^|\n)<---\n([\\s\\S]*?)\n<---(\n|$)", 2)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 5));
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(em, inner);
                parse(e, s.substring(s.indexOf("<---", r.epos) + 4));
            }
        },
        right("(^|\n)--->\n([\\s\\S]*?)\n--->(\n|$)", 2)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 5));
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(em, inner);
                parse(e, s.substring(s.indexOf("--->", r.epos) + 4));
            }
        },
        center("(^|\n)-><-\n([\\s\\S]*?)\n-><-(\n|$)", 2)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 5));
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(em, inner);
                parse(e, s.substring(s.indexOf("-><-", r.epos) + 4));
            }
        },
        code("\\{\\{\\{(.+?)\\}\\}\\}", 1)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 3));
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                em.setTextContent(recoverEscape(s.substring(r.spos, r.epos)).replace("&", "&amp;"));// double
                e.appendChild(em);
                parse(e, s.substring(r.epos + 3));
            }
        },
        css("^\\/\\*\\{\\{\\{\\*\\/\n([\\s\\S]*?)\n\\/\\*\\}\\}\\}\\*\\/$", 1)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                final Element em = e.getOwnerDocument().createElement("pre");
                em.setTextContent(s.substring(r.spos, r.epos));
                e.appendChild(em);
            }
        },
        template("^<\\!\\-\\-\\{\\{\\{\\-\\->\n([\\s\\S]*?)\n<\\!\\-\\-\\}\\}\\}\\-\\->$", 1)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                final Element em = e.getOwnerDocument().createElement("pre");
                em.setTextContent(s.substring(r.spos, r.epos));
                e.appendChild(em);
            }
        },
        comment("\\/%([\\s\\S]*?)%\\/", 1)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
            	fallback(e, s.substring(0, r.spos - 2));//↑ must be handled by text
            	//do nothing ←
            	parse(e, s.substring(r.epos + 2));//↓
            }
        },
        // avoid split paragraph
        comment_root("(^|\n)\\/%([\\s\\S]*?)%\\/(\n|$)", 2)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
            	fallback(e, s.substring(0, r.spos - 2));//↑ must be handled by text
            	//do nothing ←
            	parse(e, s.substring(r.epos + 2));//↓
            }
        },
        dd("(^|\n)([\\：]+)(.*)(\n|$)", 3)// ":" > "："
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(em, inner);
                parse(e, s.substring(r.epos));
            }
        },
        dl("(^|\n)(([;\\:]+.*(\n|$))+)", 2)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                if (e.getTagName().equals("dl"))
                {
                    int last = 0;
                    final Matcher m = pt.matcher(s);
                    while (m.find())
                    {
                        parse(e, s.substring(last, m.start(2)));
                        final String inner = m.group(2);
                        final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                        e.appendChild(em);
                        parse(em, inner.replaceAll("((^|\n)[;]+)", "$1；").replaceAll("((^|\n)[\\:]+)", "$1：").replaceAll("(^|\n)[;\\:]", "$1"));
                        last = m.end(2);
                    }
                    parse(e, s.substring(last));
                }
                else
                {
                    final String before = s.substring(0, r.spos);
                    final String after = s.substring(r.epos);
                    final String inner = s.substring(r.spos, r.epos).trim();
                    fallback(e, before);
                    final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                    e.appendChild(em);
                    parse(em, inner.replaceAll("((^|\n)[;]+)", "$1；").replaceAll("((^|\n)[\\:]+)", "$1：").replaceAll("(^|\n)[;\\:]", "$1"));
                    parse(e, after);
                }
            }
        },
        dt("(^|\n)[；]+(.*)(\n|$)", 2)// ";" > "；"
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(em, inner);
                parse(e, s.substring(r.epos));
            }
        },
        durl("(http[s]?\\:\\/\\/((?![\"' ])[a-z\\&\\?\\-\\=\\.\\/\\#\\;\\%_\\:~0-9A-Z])+)", 1)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos));
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement("a");
                em.setAttribute("href", inner);
                em.setTextContent(inner);
                e.appendChild(em);
                parse(e, s.substring(r.epos));
            }
        },
        em("\\/\\/(.+?)\\/\\/", 1)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 2));
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(em, recoverEscape(inner));
                parse(e, s.substring(s.indexOf("//", r.epos) + 2));
            }
        },
        escape("<escape[0-9]+\\/>", 0)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                final int iescbf = Singleton.INSTANCE.escbf.size();
                Singleton.INSTANCE.escbf.put("<escape" + iescbf + "/>", s.substring(r.spos, r.epos));
                parse(e, s.substring(0, r.spos) + "<escape" + iescbf + "/>" + s.substring(r.epos));
            }
        },
        h("(^|\n)(\\!+)(.*)(\n|$)", 3)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {

                final String inner = s.substring(r.spos, r.epos);
                final Matcher m = pt.matcher(s);
                m.find();
                final int depth = m.group(2).length();
                fallback(e, s.substring(0, r.spos - depth));
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement("h" + Integer.toString(depth));
                e.appendChild(em);
                parse(em, inner);
                parse(e, s.substring(r.epos));
            }
        },
        hr("(^|\n)(\\-\\-\\-\\-)(\n|$)", 2)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos));
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(e, s.substring(r.epos));
            }
        },
        img("\\[[<>]?img\\[(.*?)\\]\\]", 1) // too much duplication 2012.7.17 changed fallback and add align 2012.7.22
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
            	char alignParam = s.charAt(r.spos-5);
            	if(alignParam=='<'||alignParam=='>'){ //[>img[
            		fallback(e, s.substring(0, r.spos - 6));
            	}else{ //[img[
            		fallback(e, s.substring(0, r.spos - 5));
            	}
                
                final String inner = s.substring(r.spos, r.epos);
                
                if(inner.contains("][")){//imageLink
                	String hyperlink = inner.substring(inner.indexOf("][")+2);
                	String imginner = inner.substring(0,inner.indexOf("]["));
                	
                	//hyperLink
                	final org.w3c.dom.Element em_a = e.getOwnerDocument().createElement(Els.a.name());
                	em_a.setAttribute("href", getPropUrl(hyperlink));
                	em_a.setAttribute("style", "imageLink");
                	e.appendChild(em_a);
                	
                	//image
                    final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                    em.setAttribute("src", imginner.contains("|") ? getPropImg(imginner.split("\\|")[1]) : getPropImg(imginner));
                    em.setAttribute("alt", imginner.contains("|") ? imginner.split("\\|")[0] : imginner);
                    if(alignParam=='<') em.setAttribute("align","left");
                		else if(alignParam=='>')em.setAttribute("align","right");
                    em_a.appendChild(em);
                }else{
	                final org.w3c.dom.Element em = e.getOwnerDocument().createElement("img");
	                em.setAttribute("src", inner.contains("|") ? getPropImg(inner.split("\\|")[1]) : getPropImg(inner));
	                em.setAttribute("alt", inner.contains("|") ? inner.split("\\|")[0] : inner);
	                if(alignParam=='<') em.setAttribute("align","left");
                		else if(alignParam=='>')em.setAttribute("align","right");
	                e.appendChild(em);
                }
                parse(e, s.substring(r.epos + 2));
            }

            private String getPropImg(final String str) 
            {
                if (str.startsWith("http") || str.startsWith("ftp"))
                {
                    return str;//already url encoded
                }else if(str.startsWith("/") || str.startsWith(".")){
                	return URCF.encode(str);
                }
                return "/file/--wiki.title.encoded--/" + URCF.encode(str);//- - replaced at last
            }
            
            //copied from Els.a 2012.7.17 for imageLink
            private String getPropUrl(final String str) //being used by imageLink
            {
                if (str.startsWith("http") || str.startsWith("ftp") || str.startsWith("mms"))
                {
                    return str;//already url encoded
                }else if(str.startsWith("/") || str.startsWith(".")){
                	return URCF.encode(str);
                }
                return "/wiki/" + URCF.encode(str);
            }
        },
        li("(^|\n)([＊|＃]+)(.*)(\n|$)", 3)// "*" > "＊" "#" > "＃"
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(em, inner);
                parse(e, s.substring(r.epos));
            }
        },
        nowiki("<nowiki>([\\s\\S]*?)<\\/nowiki>", 1)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 8));
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                em.setTextContent(recoverEscape(s.substring(r.spos, r.epos)).replace("&", "&amp;"));// double
                e.appendChild(em);
                parse(e, s.substring(r.epos + 9));
            }
        },
        // avoid split paragraph
        nowiki_root("(^|\n)<nowiki>([\\s\\S]*?)<\\/nowiki>(\n|$)", 2)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 8));
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement("nowiki");
                em.setTextContent(s.substring(r.spos, r.epos).replace("&", "&amp;"));// double
                e.appendChild(em);
                parse(e, s.substring(r.epos + 9));
            }
        },
        ol("(^|\n)((#+.*(\n|$))+)", 2)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                if (e.getTagName().equals("ol"))
                {
                    int last = 0;
                    final Matcher m = pt.matcher(s);
                    while (m.find())
                    {
                        parse(e, s.substring(last, m.start(2)));
                        final String inner = m.group(2);
                        final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                        e.appendChild(em);
                        parse(em, inner.replaceAll("((^|\n)[#]+)", "$1＃").replaceAll("(^|\n)#", "$1"));
                        last = m.end(2);
                    }
                    parse(e, s.substring(last));
                }
                else
                {
                    final String before = s.substring(0, r.spos);
                    final String after = s.substring(r.epos);
                    final String inner = s.substring(r.spos, r.epos).trim();
                    fallback(e, before);
                    final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                    e.appendChild(em);
                    parse(em, inner.replaceAll("((^|\n)[#]+)", "$1＃").replaceAll("(^|\n)#", "$1"));
                    parse(e, after);
                }
            }
        },
        p("dummy", 0)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                if (s.trim().equals(""))
                {
                    return;
                }
                final String[] ps = s.trim().split("\n\n");
                for (final String p : ps)
                {
                    final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                    e.appendChild(em);
                    parse(em, p);
                }
            }
        },
        pre("(^|\n)\\{\\{\\{\n([\\s\\S]*?)\n\\}\\}\\}(\n|$)", 2)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 4));
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                em.setTextContent(s.substring(r.spos, r.epos).replace("&", "&amp;"));// double
                e.appendChild(em);
                parse(e, s.substring(r.epos + 4));
            }
        },
        root("dummy", 0)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                parse(e, s);
            }
        },
        sblockquote("(^|\n)((>+.*(\n|$))+)", 2)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                if (e.getTagName().equals("blockquote"))
                {
                    int last = 0;
                    final Matcher m = pt.matcher(s);
                    while (m.find())
                    {
                        parse(e, s.substring(last, m.start(2)));
                        final String inner = m.group(2);
                        final org.w3c.dom.Element em = e.getOwnerDocument().createElement("blockquote");
                        e.appendChild(em);
                        parse(em, inner.replaceAll("(^|\n)>", "$1"));
                        last = m.end(2);
                    }
                    parse(e, s.substring(last));
                }
                else
                {
                    final String before = s.substring(0, r.spos);
                    final String after = s.substring(r.epos);
                    final String inner = s.substring(r.spos, r.epos).trim();
                    fallback(e, before);
                    final org.w3c.dom.Element em = e.getOwnerDocument().createElement("blockquote");
                    e.appendChild(em);
                    parse(em, inner.replaceAll("(^|\n)>", "$1"));
                    parse(e, after);
                }
            }
        },
        strike("\\-\\-(.+?)\\-\\-", 1)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 2));
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(em, recoverEscape(inner));
                parse(e, s.substring(s.indexOf("--", r.epos) + 2));
            }
        },
        strong("''(.+?)''", 1)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 2));
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(em, recoverEscape(inner));
                parse(e, s.substring(s.indexOf("''", r.epos) + 2));
            }
        },
        sub("~~(.+?)~~", 1)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 2));
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(em, recoverEscape(inner));
                parse(e, s.substring(s.indexOf("~~", r.epos) + 2));
            }
        },
        sup("\\^\\^(.+?)\\^\\^", 1)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 2));
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(em, recoverEscape(inner));
                parse(e, s.substring(s.indexOf("^^", r.epos) + 2));
            }
        },
        table("(^|\n)((\\|.*\\|c?(\n|$))+)", 2)//c? caption
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos));
                final String inner = s.substring(r.spos, r.epos).trim();
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(em, inner);
                span(em);
                parse(e, s.substring(r.epos));
            }

            private void span(final org.w3c.dom.Element table)
            {
                try
                {
                    final XPath xpath = XPathFactory.newInstance().newXPath();
                    final NodeList trs = (NodeList) xpath.evaluate("//tr", table, XPathConstants.NODESET);

                    for (int k = 0; k < trs.getLength(); k++)
                    {
                        final Element n = (Element) trs.item(k);
                        int colaccum = 0;
                        final NodeList thtd = (NodeList) xpath.evaluate("th|td", n, XPathConstants.NODESET);
                        for (int i = 0; i < thtd.getLength(); i++)
                        {
                            final Element tn = (Element) thtd.item(i);
                            tn.setAttribute("col", Integer.toString(i));
                            if (!tn.getTextContent().trim().equals(">"))
                            {
                                if (colaccum != 0)
                                {
                                    tn.setAttribute("colspan", Integer.toString(colaccum + 1));
                                    colaccum = 0;
                                }
                            }
                            else
                            {
                                colaccum++;
                                tn.setAttribute("removeflag", "on");
                            }
                        }
                    }
                    for (int i = 0; ((NodeList) xpath.evaluate("//th[@col=\"" + i + "\"] | //td[@col=\"" + i + "\"]", table, XPathConstants.NODESET)).getLength() > 0; i++)
                    {
                        final NodeList row = (NodeList) xpath.evaluate("//th[@col=\"" + i + "\"] | //td[@col=\"" + i + "\"]", table, XPathConstants.NODESET);
                        int rowaccum = 0;
                        for (int r = row.getLength() - 1; r >= 0; r--)
                        {
                            final Element oe = (Element) row.item(r);
                            if (!oe.getTextContent().trim().equals("~"))
                            {
                                if (rowaccum != 0)
                                {
                                    oe.setAttribute("rowspan", Integer.toString(rowaccum + 1));
                                    rowaccum = 0;
                                }
                            }
                            else
                            {
                                rowaccum++;
                                oe.setAttribute("removeflag", "on");
                            }
                        }
                    }
                    final NodeList deln = (NodeList) xpath.evaluate("//th[@removeflag=\"on\"]|//td[@removeflag=\"on\"]", table, XPathConstants.NODESET);
                    for (int i = 0; i < deln.getLength(); i++)
                    {
                        deln.item(i).getParentNode().removeChild(deln.item(i));
                    }
                    final NodeList tcs = (NodeList) xpath.evaluate("//th|//td", table, XPathConstants.NODESET);
                    for (int i = 0; i < tcs.getLength(); i++)
                    {
                        ((Element) tcs.item(i)).removeAttribute("col");
                    }
                }
                catch (final XPathExpressionException e)
                {
                    e.printStackTrace();
                }
            }
        },
        nolinetable("(^|\n)\\-((\\|.*\\|c?(\n|$))+)", 2)//c? caption 2012.7.16
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos-1));
                final String inner = s.substring(r.spos, r.epos).trim();
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement("table");
                em.setAttribute("class", "noline");
                e.appendChild(em);
                parse(em, inner);
                span(em);
                parse(e, s.substring(r.epos));
            }

            private void span(final org.w3c.dom.Element table)
            {
                try
                {
                    final XPath xpath = XPathFactory.newInstance().newXPath();
                    final NodeList trs = (NodeList) xpath.evaluate("//tr", table, XPathConstants.NODESET);

                    for (int k = 0; k < trs.getLength(); k++)
                    {
                        final Element n = (Element) trs.item(k);
                        int colaccum = 0;
                        final NodeList thtd = (NodeList) xpath.evaluate("th|td", n, XPathConstants.NODESET);
                        for (int i = 0; i < thtd.getLength(); i++)
                        {
                            final Element tn = (Element) thtd.item(i);
                            tn.setAttribute("col", Integer.toString(i));
                            if (!tn.getTextContent().trim().equals(">"))
                            {
                                if (colaccum != 0)
                                {
                                    tn.setAttribute("colspan", Integer.toString(colaccum + 1));
                                    colaccum = 0;
                                }
                            }
                            else
                            {
                                colaccum++;
                                tn.setAttribute("removeflag", "on");
                            }
                        }
                    }
                    for (int i = 0; ((NodeList) xpath.evaluate("//th[@col=\"" + i + "\"] | //td[@col=\"" + i + "\"]", table, XPathConstants.NODESET)).getLength() > 0; i++)
                    {
                        final NodeList row = (NodeList) xpath.evaluate("//th[@col=\"" + i + "\"] | //td[@col=\"" + i + "\"]", table, XPathConstants.NODESET);
                        int rowaccum = 0;
                        for (int r = row.getLength() - 1; r >= 0; r--)
                        {
                            final Element oe = (Element) row.item(r);
                            if (!oe.getTextContent().trim().equals("~"))
                            {
                                if (rowaccum != 0)
                                {
                                    oe.setAttribute("rowspan", Integer.toString(rowaccum + 1));
                                    rowaccum = 0;
                                }
                            }
                            else
                            {
                                rowaccum++;
                                oe.setAttribute("removeflag", "on");
                            }
                        }
                    }
                    final NodeList deln = (NodeList) xpath.evaluate("//th[@removeflag=\"on\"]|//td[@removeflag=\"on\"]", table, XPathConstants.NODESET);
                    for (int i = 0; i < deln.getLength(); i++)
                    {
                        deln.item(i).getParentNode().removeChild(deln.item(i));
                    }
                    final NodeList tcs = (NodeList) xpath.evaluate("//th|//td", table, XPathConstants.NODESET);
                    for (int i = 0; i < tcs.getLength(); i++)
                    {
                        ((Element) tcs.item(i)).removeAttribute("col");
                    }
                }
                catch (final XPathExpressionException e)
                {
                    e.printStackTrace();
                }
            }
        },
        tabtable("(^|\n)\\[\\[\\[\n([\\s\\S]*?)\n\\]\\]\\](\n|$)", 2)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos-4));
                final String inner = "|"+s.substring(r.spos, r.epos).trim().replace("|","<nowiki>|</nowiki>").replace("\t", "|").replace("\n","|\n|")+"|";
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement("table");
                e.appendChild(em);
                parse(em, inner);
                span(em);
                parse(e, s.substring(s.indexOf("]]]", r.epos) + 3));
            }

            private void span(final org.w3c.dom.Element table)
            {
                try
                {
                    final XPath xpath = XPathFactory.newInstance().newXPath();
                    final NodeList trs = (NodeList) xpath.evaluate("//tr", table, XPathConstants.NODESET);

                    for (int k = 0; k < trs.getLength(); k++)
                    {
                        final Element n = (Element) trs.item(k);
                        int colaccum = 0;
                        final NodeList thtd = (NodeList) xpath.evaluate("th|td", n, XPathConstants.NODESET);
                        for (int i = 0; i < thtd.getLength(); i++)
                        {
                            final Element tn = (Element) thtd.item(i);
                            tn.setAttribute("col", Integer.toString(i));
                            if (!tn.getTextContent().trim().equals(">"))
                            {
                                if (colaccum != 0)
                                {
                                    tn.setAttribute("colspan", Integer.toString(colaccum + 1));
                                    colaccum = 0;
                                }
                            }
                            else
                            {
                                colaccum++;
                                tn.setAttribute("removeflag", "on");
                            }
                        }
                    }
                    for (int i = 0; ((NodeList) xpath.evaluate("//th[@col=\"" + i + "\"] | //td[@col=\"" + i + "\"]", table, XPathConstants.NODESET)).getLength() > 0; i++)
                    {
                        final NodeList row = (NodeList) xpath.evaluate("//th[@col=\"" + i + "\"] | //td[@col=\"" + i + "\"]", table, XPathConstants.NODESET);
                        int rowaccum = 0;
                        for (int r = row.getLength() - 1; r >= 0; r--)
                        {
                            final Element oe = (Element) row.item(r);
                            if (!oe.getTextContent().trim().equals("~"))
                            {
                                if (rowaccum != 0)
                                {
                                    oe.setAttribute("rowspan", Integer.toString(rowaccum + 1));
                                    rowaccum = 0;
                                }
                            }
                            else
                            {
                                rowaccum++;
                                oe.setAttribute("removeflag", "on");
                            }
                        }
                    }
                    final NodeList deln = (NodeList) xpath.evaluate("//th[@removeflag=\"on\"]|//td[@removeflag=\"on\"]", table, XPathConstants.NODESET);
                    for (int i = 0; i < deln.getLength(); i++)
                    {
                        deln.item(i).getParentNode().removeChild(deln.item(i));
                    }
                    final NodeList tcs = (NodeList) xpath.evaluate("//th|//td", table, XPathConstants.NODESET);
                    for (int i = 0; i < tcs.getLength(); i++)
                    {
                        ((Element) tcs.item(i)).removeAttribute("col");
                    }
                }
                catch (final XPathExpressionException e)
                {
                    e.printStackTrace();
                }
            }
        },

        tc("\\|(.*?)(\\||$)", 1)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                final String cell = s.substring(r.spos, r.epos);

                String align;
                if ((cell.startsWith(" ") && cell.endsWith(" ")) | (!cell.startsWith(" ") && !cell.endsWith(" ")))
                {
                    align = "center";
                }
                else if (cell.startsWith(" ") && !cell.endsWith(" "))
                {
                    align = "right";
                }
                else
                {
                    align = "left";
                }
                final org.w3c.dom.Element em = (cell.trim().startsWith("!")) ? e.getOwnerDocument().createElement("th") : e.getOwnerDocument().createElement("td");
                final String inner = cell.trim().startsWith("!") ? cell.replaceFirst("\\!", "") : cell;
                em.setAttribute("align", align);
                e.appendChild(em);

                parse(em, recoverEscape(inner));
                parse(e, s.substring(r.epos));
            }
        },
        text("dummy", 0)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                final Matcher m = escape.pt.matcher(s);
                if (m.find())
                {
                    handle(e, s.substring(0, m.start()), null);// ⇑
                    parse(e, recoverEscape(s.substring(m.start())));// ↲
                    return;
                }// ⇓

                if (s.trim().equals("") && !e.getTagName().equals("p"))
                {
                    return;
                }
                for (int i = 0; i < s.length(); i++)
                {
                    if (s.charAt(i) != '\n')
                    {
                        sb.append(s.charAt(i));
                    }
                    else
                    {
                        e.appendChild(e.getOwnerDocument().createTextNode(sb.toString()));
                        e.appendChild(e.getOwnerDocument().createElement("br"));
                        sb.delete(0, sb.length());
                    }
                }
                if (sb.length() > 0)
                {
                    e.appendChild(e.getOwnerDocument().createTextNode(sb.toString()));
                    sb.delete(0, sb.length());
                }
            }

            private final StringBuffer sb = new StringBuffer();
        },
        tr("(^|\n)(\\|.*)\\|(\n|$)", 2)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(em, inner);
                parse(e, s.substring(r.epos + 1));
            }
        },
        caption("(^|\n)\\|(.*)\\|c(\n|$)", 2)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(em, inner);
                parse(e, s.substring(r.epos + 2));//length(|,c)
            }
        },
        u("__(.+?)__", 1)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 2));
                final String inner = s.substring(r.spos, r.epos);
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                e.appendChild(em);
                parse(em, recoverEscape(inner));
                parse(e, s.substring(s.indexOf("__", r.epos) + 2));
            }
        },
        ul("(^|\n)((\\*+.*(\n|$))+)", 2)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                if (e.getTagName().equals("ul"))
                {
                    int last = 0;
                    final Matcher m = pt.matcher(s);
                    while (m.find())
                    {
                        parse(e, s.substring(last, m.start(2)));
                        final String inner = m.group(2);
                        final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                        e.appendChild(em);
                        parse(em, inner.replaceAll("((^|\n)[\\*]+)", "$1＊").replaceAll("(^|\n)\\*", "$1"));
                        last = m.end(2);
                    }
                    parse(e, s.substring(last));
                }
                else
                {
                    final String inner = s.substring(r.spos, r.epos).trim();
                    fallback(e, s.substring(0, r.spos));// ⇑
                    final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                    e.appendChild(em);
                    parse(em, inner.replaceAll("((^|\n)[\\*]+)", "$1＊").replaceAll("(^|\n)\\*", "$1"));
                    parse(e, s.substring(r.epos));// ⇓
                }
            }
        },
        userhtml("<html>([\\s\\S]*?)<\\/html>", 1)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 6));
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement(name());
                em.setTextContent(recoverEscape(s.substring(r.spos, r.epos)));
                e.appendChild(em);
                parse(e, s.substring(r.epos + 7));
            }
        },
        userhtml_root("(^|\n)<html>([\\s\\S]*?)<\\/html>(\n|$)", 2)
        {
            @Override
            public void handle(final org.w3c.dom.Element e, final String s, final Range r)
            {
                fallback(e, s.substring(0, r.spos - 6));
                final org.w3c.dom.Element em = e.getOwnerDocument().createElement("userhtml");
                em.setTextContent(s.substring(r.spos, r.epos));
                e.appendChild(em);
                parse(e, s.substring(r.epos + 7));
            }
        };

        private final int gr;
        protected final Pattern pt;
        
        Els(final String match, final int group)
        {
            this.pt = Pattern.compile(match);
            this.gr = group;
        }
        
        public void handle(Element e, String s, Range r) {}

		public void fallback(final org.w3c.dom.Element e, final String s)
        {
            if (e.getTagName().equals("div") | e.getTagName().equals("blockquote") | e.getTagName().equals("left") | e.getTagName().equals("right") | e.getTagName().equals("center")) p.handle(e, s, null);
            else text.handle(e, s, null);
        }
        
		public Range[] match(final String s)
        {
            final Matcher m = pt.matcher(s);
            if (m.find())
            {
                return new Range[] { new Range(m.start(), m.end()), new Range(m.start(this.gr), m.end(this.gr)) };
            }
            return null;
        }

        public void parse(final org.w3c.dom.Element re, final String s)
        {
            final Singleton.Index idx = Singleton.INSTANCE.idx;
            idx.imap().clear();
            idx.rmap().clear();
            for (final El el : chn.get(re.getTagName()))
            {
                final Range[] rs = el.match(s);
                if (rs != null)
                {
                    idx.imap().put(rs[0], el);
                    idx.rmap().put(el, rs[1]);
                }
            }
            final El best = idx.best();
            if (best != null) best.handle(re, s, idx.rmap().get(best));
            else fallback(re, s);
        }

        public String recoverEscape(final String s)
        {
            if (!escape.pt.matcher(s).find()) return s;
            if (Singleton.INSTANCE.escbf.size() == 0) return s;
            String tmp = s;
            for (int i = Singleton.INSTANCE.escbf.size() - 1; i >= 0; i--)
            {
                tmp = tmp.replace("<escape" + Integer.toString(i) + "/>", Singleton.INSTANCE.escbf.get("<escape" + Integer.toString(i) + "/>"));
            }
            return tmp;
        }
        
        public enum Singleton
        {
            INSTANCE;
            
            private final Map<String, String> escbf = Collections.synchronizedMap(new HashMap<String, String>());
            private final Index idx = new Index();
            
            class Index
            {
                public El best()
                {
                    cross.clear();
                    final Set<Range> rs = imap.keySet();
                    Range bsRange = null;
                    for (final Range r : rs)
                    {
                        if (bsRange == null) bsRange = r;
                        else if (r.spos < bsRange.spos) bsRange = r;
                    }
                    for (final Range r : rs)
                    {
                        if (bsRange == null) bsRange = r;
                        else if ((r.spos < bsRange.spos) && (bsRange.epos < r.epos)) ;
                        else if ((bsRange.spos < r.spos) && (r.epos < bsRange.epos)) ;
                        else if ((r.spos < bsRange.spos) && (bsRange.spos < r.epos)) cross.add(r);
                        else if ((bsRange.spos < r.spos) && (r.spos < bsRange.epos)) cross.add(r);
                    }
                    if ((cross.size() > 0) && (bsRange != null))
                    {
                        Collections.sort(cross, new Comparator<Range>()
                        {
                            public int compare(final Range r0, final Range r1)
                            {
                                if (r0.spos > r1.spos) return +1;
                                else if (r0.spos < r1.spos) return -1;
                                else return 0;
                            }
                        });
                        final El original = imap.get(bsRange);
                        final Range escapeRange = cross.get(0);
                        final El escel = imap.get(escapeRange);
                        if (nowiki.equals(escel) | userhtml.equals(escel) | code.equals(escel))
                        {
                            rmap.put(escape, escapeRange);
                            return escape;
                        }
                        if (original.equals(em) | original.equals(tc))
                        {
                            rmap.put(escape, escapeRange);
                            return escape;
                        }
                    }
                    if (imap.size() != 0){}
                    return imap.get(bsRange);
                }

                public Map<Range, El> imap() {return this.imap;}
                public Map<El, Range> rmap() {return this.rmap;}
                private final List<Range> cross = Collections.synchronizedList(new ArrayList<Range>());
                private final Map<Range, El> imap = Collections.synchronizedMap(new HashMap<Range, El>());
                private final Map<El, Range> rmap = Collections.synchronizedMap(new HashMap<El, Range>());
            }
        }

        class Range
        {
            public Range(final int spos, final int epos)
            {
                this.spos = spos;
                this.epos = epos;
            }
            protected final int epos;
            protected final int spos;
        }
    }

   

    
    /**
     * @param s
     *            wiki text
     * @param wiki 
     * 			  wiki title
     * @return html
     */
    public String parse(final String s, String wiki_title_encoded)
    {
        if(s.trim().equals(""))	return "";
        
        final Document doc = DBF.get().newDocument();
        final org.w3c.dom.Element re = doc.createElement("div");
        doc.appendChild(re);
        Els.root.handle(re, s, null);
        final Matcher m = PTUSERHTML.matcher(d2s(doc).replace("&amp;", "&"));
        final StringBuffer sb = new StringBuffer();
        while (m.find())
        {
            m.appendReplacement(sb, Miscellaneous.INSTANCE.htmlDecode(m.group(0)));
        }
        m.appendTail(sb);
        return sb.toString().replace("--wiki.title.encoded--", wiki_title_encoded);
    }
    
    public String d2s(final Document doc)
    {
        final StringWriter sw = new StringWriter();
        final StreamResult result = new StreamResult(sw);
        try
        {
            TFF.get().transform(new DOMSource(doc), result);
            return sw.toString();
        }
        catch (final TransformerException e)
        {
            e.printStackTrace();
        }
        return null;
    }
    
    private static final Pattern PTUSERHTML = Pattern.compile("<userhtml>([\\s\\S]*?)</userhtml>");
    public final static HashMap<String, Els[]> chn = new HashMap<String, Els[]>();
    {// parse flow
        final Els[] al = new Els[] { Els.a, Els.code, Els.durl, Els.em, Els.img, Els.nowiki, Els.strike, Els.strong, Els.sub, Els.sup, Els.u, Els.userhtml, Els.comment };
        final Els[] ne = new Els[] {};
        final Els[] rt = new Els[] { Els.blockquote, Els.left, Els.right, Els.center, Els.css, Els.template, Els.dl, Els.h, Els.hr, Els.nowiki_root, Els.ol, Els.pre, Els.sblockquote, Els.table, Els.nolinetable, Els.tabtable, Els.ul, Els.userhtml_root, Els.comment_root };
        final Els[] ts = new Els[] { Els.text };
        final Els[] tr = new Els[] { Els.tr, Els.caption };
        final Els[] ds = new Els[] { Els.dl, Els.dd, Els.dt };
        final Els[] tc = new Els[] { Els.tc, Els.a, Els.code, Els.durl, Els.em, Els.img, Els.nowiki, Els.strike, Els.strong, Els.sub, Els.sup, Els.u, Els.userhtml, Els.comment };
        final Els[] ol = new Els[] { Els.ol, Els.li };
        final Els[] ul = new Els[] { Els.ul, Els.li };
        chn.put("a", ts);
        chn.put("blockquote", rt);
        chn.put("left", rt);
        chn.put("right", rt);
        chn.put("center", rt);
        chn.put("code", ts);
        chn.put("dd", al);
        chn.put("dt", al);
        chn.put("durl", ts);
        chn.put("em", al);
        chn.put("h1", al);
        chn.put("h2", al);
        chn.put("h3", al);
        chn.put("h4", al);
        chn.put("h5", al);
        chn.put("img", ts);
        chn.put("nowiki", ts);
        chn.put("nowiki_root", ts);
        chn.put("ol", ol);
        chn.put("p", al);
        chn.put("pre", ts);
        chn.put("div", rt);
        chn.put("sblockquote", al);
        chn.put("strike", al);
        chn.put("strong", al);
        chn.put("sub", al);
        chn.put("sup", al);
        chn.put("text", ne);
        chn.put("tr", tc);
        chn.put("u", al);
        chn.put("ul", ul);
        chn.put("userhtml", ts);
        chn.put("userhtml_root", ts);
        chn.put("li", al);
        chn.put("table", tr);
        chn.put("dl", ds);
        chn.put("hr", ne);
        chn.put("td", al);
        chn.put("th", al);
        chn.put("caption", al);
    }
}
